/*
 * kms.cpp
 *
 * @author Pietro Spinella (scrockman[at]gmail[dot]com)
 * @license GPLv3
 */

#include <iostream>
#include <QByteArray>
#include <QDomElement>
#include <QDomNode>
#include <QFile>
#include <QHash>
#include <QUrl>

#include "kms.h"
#include "defaultiostrategy.h"

using namespace std;
using namespace kms;

Kms::Kms() {

	// Creates the enhanced qhttp and connects the "request performed" signal of _enhancedQHttp with the "request performed" slot of this
	_enhancedQHttp = new EnhancedQHttp;
	connect(_enhancedQHttp, SIGNAL(requestPerformed(QByteArray &)), this, SLOT(requestPerformed(QByteArray &)));

	// Creates the default io strategy
	_ioStrategy = new DefaultIOStrategy;

	// Creates the input list null object and the used input list one
	_inputListNullObject = new QList<Input *>;
	_usedInputListNullObject = new QList<UsedInput *>;

	// Creates the tree
	_kmsRoot = new KmsRoot;

	// Sets the running flag to false
	_runningFlag = false;

}

Kms::~Kms() {

	// Disconnects all the slots connected with _enhancedQHttp and deletes _enhancedQHttp
	_enhancedQHttp->disconnect();
	delete _enhancedQHttp;

	// Deletes the io strategy
	delete _ioStrategy;

	// Deletes the input list null object and the used input list one
	delete _inputListNullObject;
	delete _usedInputListNullObject;

	// Deletes the tree
	delete _kmsRoot;

}

void Kms::setIOStrategy(const IOStrategy *ioStrategy) {

	// Deletes the old io strategy
	delete _ioStrategy;

	// Sets the new io strategy
	_ioStrategy = (IOStrategy *) ioStrategy;

}

bool Kms::load(const QString &fileName) {

	// deletes and creates the tree
	delete _kmsRoot;
	_kmsRoot = new KmsRoot;

	// If kms is running then return the failure...
	if(_runningFlag == true) {
		return false;
	}

	// Opens the xml file
	QFile qFile(fileName);

	// If the xml file has not been opened...
	if(qFile.open(QIODevice::ReadOnly) == false) {
		qDebug("[kms::Kms] ERROR: can't open the XML file!");
		return false;
	}

	// Creates and sets the dom document
	QDomDocument qDomDocument;
	QString errorMsg;
	int errorLine;
	int errorColumn;
	bool okFlag = qDomDocument.setContent(&qFile, &errorMsg, &errorLine, &errorColumn);

	// Closes the xml file
	qFile.close();

	// If the dom document has not been set...
	if(okFlag == false) {
		qDebug("[kms::Kms] ERROR: can't set the DOM document!");
		cout << QString("  " + errorMsg + " at row " + QString("%1").arg(errorLine) + ", column " + QString("%1").arg(errorColumn)).toStdString() << endl;
		return false;
	}

	// Fills the tree
	okFlag = _fillTree(&qDomDocument);

	// If the tree has not been set...
	if(okFlag == false) {
		return false;
	}

	// Returns the success
	return true;

}

const QList<Input *> *Kms::getInputList(const QString &serviceId) const {

	// Returns the input list
	const QList<Service *> *serviceList = _kmsRoot->getServices()->getServiceList();
	int serviceListSize = serviceList->size();
	for(int i = 0; i < serviceListSize; i++) {
		Service *service = serviceList->at(i);
		if(service->getId() == serviceId) {
			return service->getInputs()->getInputList();
		}
	}

	// Returns the input list null object
	return _inputListNullObject;

}

const QList<UsedInput *> *Kms::getUsedInputList(const QString &serviceId, const QString &sequenceId) const {

	// Returns the used input list
	const QList<Service *> *serviceList = _kmsRoot->getServices()->getServiceList();
	int serviceListSize = serviceList->size();
	for(int i = 0; i < serviceListSize; i++) {
		Service *service = serviceList->at(i);
		if(service->getId() == serviceId) {
			const QList<Sequence *> *sequenceList = service->getSequences()->getSequenceList();
			int sequenceListSize = sequenceList->size();
			for(int j = 0; j < sequenceListSize; j++) {
				Sequence *sequence = sequenceList->at(j);
				if(sequence->getId() == sequenceId) {
					return sequence->getUsedInputs()->getUsedInputList();
				}
			}
		}
	}

	// Returns the used input list null object
	return _usedInputListNullObject;

}

int Kms::getInputCount(const QString &serviceId, const QString &type) const {

	// Gets the input list and its size
	const QList<Input *> *inputList = getInputList(serviceId);
	int inputListSize = inputList->size();

	// If the user wants all the inputs...
	if(type == 0) {
		return inputListSize;
	}

	// Stores the result
	int result = 0;

	// Evaluates the input count
	for(int i = 0; i < inputListSize; i++) {
		if(inputList->at(i)->getType() == type) {
			result++;
		}
	}

	// Returns the result
	return result;

}

int Kms::getUsedInputCount(const QString &serviceId, const QString &sequenceId, const QString &type) const {

	// Gets the used input list and its size
	const QList<UsedInput *> *usedInputList = getUsedInputList(serviceId, sequenceId);
	int usedInputListSize = usedInputList->size();

	// If the user wants all the inputs...
	if(type == 0) {
		return usedInputListSize;
	}

	// Stores the result
	int result = 0;

	// Gets the used input list and its size
	const QList<Input *> *inputList = getInputList(serviceId);
	int inputListSize = inputList->size();

	// Evaluates the used input count
	for(int i = 0; i < usedInputListSize; i++) {
		UsedInput *usedInput = usedInputList->at(i);
		for(int j = 0; j < inputListSize; j++) {
			Input *input = inputList->at(j);
			if(usedInput->getInput() == input->getId()) {
				if(input->getType() == type) {
					result++;
				}
			}
		}
	}

	// Returns the result
	return result;

}

const QString Kms::evaluateText(const QString &serviceId, const QString &string, const QString &language) {

	// Stores the result
	QString result = string;

	// Replaces all the occurrences of "\$TEXT\[.*\]"...
	while(QRegExp("(.*)(\\$TEXT\\[(.*)\\])(.*)").exactMatch(result) == true) {
		int index1 = result.indexOf(QRegExp("(\\$TEXT\\[(.*)\\])"));
		if(index1 == -1) {
			break;
		}
		index1 += 6;
		int index2 = result.indexOf(QRegExp("(\\])"), index1);
		QString current = result;
		current.chop(current.size() - index2);
		current.remove(0, index1);
		result = _pickText(serviceId, current, language);
	}

	// Returns the result
	return result;

}


bool Kms::performSequence(const QString &serviceId, const QString &sequenceId) const {

	// Wraps the other performSequence method
	QList< QPair<QString, QString> > inputDataList;
	return performSequence(serviceId, sequenceId, inputDataList);

}

bool Kms::performSequence(const QString &serviceId, const QString &sequenceId, const QList< QPair<QString, QString> > &inputDataList) const {

//? QHash<QString, QString> hash;
//? QByteArray data;
//? _ioStrategy->getCaptchaInput(hash, data);

	// Checks if the engine is still running
	if(_runningFlag == true) {
		return false;
	}

	// Sets the running flag true
	((bool &) _runningFlag) = true;

	// Clears the current http request data list
	((QList< QPair<QString, QByteArray> >) _currentHttpRequestDataList).clear();

	// Does the real performSequence method
	bool result = _performSequence(serviceId, sequenceId, inputDataList);

	// Sets the running flag false
	((bool &) _runningFlag) = false;

	// Emits the sequenceFinished signal
// TODO	emit ((Kms*) this)->sequenceFinished();

	// Returns the result
	return result;

}

bool Kms::_fillTree(const QDomDocument *qDomDocument) {

	// <kmsRoot>
	QDomElement e = qDomDocument->documentElement();
	if(e.isNull() == true || e.tagName() != "kmsRoot") {
		qDebug("[kms::Kms] ERROR: <kmsRoot> element not found!");
		return false;
	}

	// Checks the right version
	if(e.attribute("version","0.1") != "0.1") {
		qDebug("[kms::Kms] ERROR: <kmsRoot> version not supported (!= 0.1)!");
		return false;
	}

	// Sets the version
	_kmsRoot->setVersion(e.attribute("version", "0.1"));

	// For each <kmsRoot> child (<texts> | <services>)
	bool textsFlag = false;
	bool servicesFlag = false;
	Texts *texts = (Texts *) _kmsRoot->getTexts();
	Services *services = (Services *) _kmsRoot->getServices();
	for(QDomNode n = e.firstChild(); n.isNull() == false; n = n.nextSibling()) {

		// Gets the element (<texts> | <services>)...
		e = n.toElement();

		// If the element is the first occurrence of <texts>...
		if(textsFlag == false && e.isNull() == false && e.tagName() == "texts") {

			// First occurrence of <texts> element found
			textsFlag = true;

			// Sets the default language
			texts->setDefaultLanguage(e.attribute("defaultLanguage", "en"));

			// For each <texts> child (<text>)
			QList<Text *> *textList = (QList<Text *> *) texts->getTextList();
			for(QDomNode n2 = n.firstChild(); n2.isNull() == false; n2 = n2.nextSibling()) {

				// <text>
				e = n2.toElement();
				if(e.isNull() == false && e.tagName() == "text") {

					// Sets the attributes
					Text *text = new Text;
					text->setId(e.attribute("id", ""));
					textList->append(text);

					// For each <text> child (<value>)
					QList<Value *> *valueList = (QList<Value *> *) text->getValueList();
					for(QDomNode n3 = n2.firstChild(); n3.isNull() == false; n3 = n3.nextSibling()) {

						// <value>
						e = n3.toElement();
						if(e.isNull() == false && e.tagName() == "value") {

							// Sets the attributes
							Value *value = new Value;
							value->setLanguage(e.attribute("language", ""));
							value->setValue(e.attribute("value", ""));
							valueList->append(value);

						}

					}

				}

			}

		}

		// If the element is the first occurrence of <services>...
		if(servicesFlag == false && e.isNull() == false && e.tagName() == "services") {

			// First occurrence of <services> element found
			servicesFlag = true;

			// For each <services> child (<service>)
			QList<Service *> *serviceList = (QList<Service *> *) services->getServiceList();
			for(QDomNode n2 = n.firstChild(); n2.isNull() == false; n2 = n2.nextSibling()) {

				// <service>
				e = n2.toElement();
				if(e.isNull() == false && e.tagName() == "service") {

					// Sets the attributes
					Service *service = new Service;
					service->setId(e.attribute("id", ""));
					service->setName(e.attribute("name", ""));
					service->setIcon(e.attribute("icon", ""));
					service->setDescription(e.attribute("description", ""));
					service->setVersion(e.attribute("version", ""));
					service->setAuthor(e.attribute("author", ""));
					service->setEnabled(e.attribute("enabled", "true"));
					serviceList->append(service);

					// For each <services> child (<texts> | <inputs> | <outputs> | <actions> | <sequences>)
					bool textsFlag = false;
					bool inputsFlag = false;
					bool outputsFlag = false;
					bool actionsFlag = false;
					bool sequencesFlag = false;
					Texts *texts = (Texts *) service->getTexts();
					Inputs *inputs = (Inputs *) service->getInputs();
					Outputs *outputs = (Outputs *) service->getOutputs();
					Actions *actions = (Actions *) service->getActions();
					Sequences *sequences = (Sequences *) service->getSequences();
					for(QDomNode n3 = n2.firstChild(); n3.isNull() == false; n3 = n3.nextSibling()) {

						// Gets the element (<texts> | <inputs> | <outputs> | <actions> | <sequences>)...
						e = n3.toElement();

						// If the element is the first occurrence of <texts>...
						if(textsFlag == false && e.isNull() == false && e.tagName() == "texts") {

							// First occurrence of <texts> element found
							textsFlag = true;

							// Sets the default language
							texts->setDefaultLanguage(e.attribute("defaultLanguage", "en"));

							// For each <texts> child (<text>)
							QList<Text *> *textList = (QList<Text *> *) texts->getTextList();
							for(QDomNode n4 = n3.firstChild(); n4.isNull() == false; n4 = n4.nextSibling()) {

								// <text>
								e = n4.toElement();
								if(e.isNull() == false && e.tagName() == "text") {

									// Sets the attributes
									Text *text = new Text;
									text->setId(e.attribute("id", ""));
									textList->append(text);

									// For each <text> child (<value>)
									QList<Value *> *valueList = (QList<Value *> *) text->getValueList();
									for(QDomNode n5 = n4.firstChild(); n5.isNull() == false; n5 = n5.nextSibling()) {

										// <value>
										e = n5.toElement();
										if(e.isNull() == false && e.tagName() == "value") {

											// Sets the attributes
											Value *value = new Value;
											value->setLanguage(e.attribute("language", ""));
											value->setValue(e.attribute("value", ""));
											valueList->append(value);

										}

									}

								}

							}

						}

						// If the element is the first occurrence of <inputs>...
						if(inputsFlag == false && e.isNull() == false && e.tagName() == "inputs") {

							// First occurrence of <inputs> element found
							inputsFlag = true;

							// For each <inputs> child (<input>)
							QList<Input *> *inputList = (QList<Input *> *) inputs->getInputList();
							for(QDomNode n4 = n3.firstChild(); n4.isNull() == false; n4 = n4.nextSibling()) {

								// <input>
								e = n4.toElement();
								if(e.isNull() == false && e.tagName() == "input") {

									// Sets the attributes
									Input *input = new Input;
									input->setId(e.attribute("id", ""));
									input->setContent(e.attribute("content", ""));
									input->setType(e.attribute("type", ""));
									input->setRegex(e.attribute("regex", ".*"));
									input->setName(e.attribute("name", ""));
									input->setDefaultValue(e.attribute("defaultValue", ""));
									inputList->append(input);

								}

							}

						}

						// If the element is the first occurrence of <outputs>...
						if(outputsFlag == false && e.isNull() == false && e.tagName() == "outputs") {

							// First occurrence of <outputs> element found
							outputsFlag = true;

							// For each <outputs> child (<output>)
							QList<Output *> *outputList = (QList<Output *> *) outputs->getOutputList();
							for(QDomNode n4 = n3.firstChild(); n4.isNull() == false; n4 = n4.nextSibling()) {

								// <output>
								e = n4.toElement();
								if(e.isNull() == false && e.tagName() == "output") {

									// Sets the attributes
									Output *output = new Output;
									output->setId(e.attribute("id", ""));
									output->setContent(e.attribute("content", "string"));
									output->setSource(e.attribute("source", ""));
									outputList->append(output);

								}

							}

						}

						// If the element is the first occurrence of <actions>...
						if(actionsFlag == false && e.isNull() == false && e.tagName() == "actions") {

							// First occurrence of <actions> element found
							actionsFlag = true;

							// For each <actions> child (<action>)
							QList<Action *> *actionList = (QList<Action *> *) actions->getActionList();
							for(QDomNode n4 = n3.firstChild(); n4.isNull() == false; n4 = n4.nextSibling()) {

								// <action>
								e = n4.toElement();
								if(e.isNull() == false && e.tagName() == "action") {

									// Creates the action
									Action *action = new Action;
									action->setId(e.attribute("id", ""));
									action->setType(e.attribute("type", ""));

									// Checks if the internal node of >action type="$type"> is <$type>, if not don't insert the current action into the tree
									QString type = e.attribute("type", "");
									QDomNode n5 = n4.firstChild();
									if(n5.isNull() == true) {
										continue;
									}
									e = n5.toElement();
									if(e.isNull() == true) {
										continue;
									}
									if(e.tagName() != type) {
										continue;
									}

									// Case: <httpRequest>
									if(type == "httpRequest") {
										// Creates and sets the "httpRequest"
										HttpRequest *httpRequest = new HttpRequest;
										httpRequest->setId(e.attribute("id", ""));
										httpRequest->setText(e.text());
										action->setHttpRequest(httpRequest);
										actionList->append(action);
										// Inserts the (httpRequestId, httpRequestData) pair into the current http request data list
										QPair<QString, QByteArray> pair(httpRequest->getId(), QByteArray());
										_currentHttpRequestDataList.append(pair);
									}

									// Case: <checkHttpRequest>
									if(type == "checkHttpRequest") {
										// Creates and sets the "checkHttpRequest"
										CheckHttpRequest *checkHttpRequest = new CheckHttpRequest;
										checkHttpRequest->setHttpRequest(e.attribute("httpRequest", ""));
										checkHttpRequest->setContainsRegex(e.attribute("containsRegex", ""));
										checkHttpRequest->setOnTrueOutput(e.attribute("onTrueOutput", ""));
										checkHttpRequest->setOnFalseOutput(e.attribute("onFalseOutput", ""));
										checkHttpRequest->setOnTrue(e.attribute("onTrue", ""));
										checkHttpRequest->setOnFalse(e.attribute("onFalse", ""));
										action->setCheckHttpRequest(checkHttpRequest);
										actionList->append(action);
									}

									// Case: <askCaptchaCode>
									if(type == "askCaptchaCode") {
										// Creates and sets the "askCaptchaCode"
										AskCaptchaCode *askCaptchaCode = new AskCaptchaCode;
										askCaptchaCode->setHttpRequest(e.attribute("httpRequest", ""));
										askCaptchaCode->setInput(e.attribute("input", ""));
										action->setAskCaptchaCode(askCaptchaCode);
										actionList->append(action);
									}

								}

							}

						}

						// If the element is the first occurrence of <sequences>...
						if(sequencesFlag == false && e.isNull() == false && e.tagName() == "sequences") {

							// First occurrence of <sequences> element found
							sequencesFlag = true;

							// For each <sequences> child (<sequence>)
							QList<Sequence *> *sequenceList = (QList<Sequence *> *) sequences->getSequenceList();
							for(QDomNode n4 = n3.firstChild(); n4.isNull() == false; n4 = n4.nextSibling()) {

								// <sequence>
								e = n4.toElement();
								if(e.isNull() == false && e.tagName() == "sequence") {

									// Sets the attributes
									Sequence *sequence = new Sequence;
									sequence->setId(e.attribute("id", ""));
									sequenceList->append(sequence);

									// For each <sequence> child (<usedInputs> | <usedOutputs> | <usedActions>)
									bool usedInputsFlag = false;
									bool usedActionsFlag = false;
									UsedInputs *usedInputs = (UsedInputs *) sequence->getUsedInputs();
									UsedActions *usedActions = (UsedActions *) sequence->getUsedActions();
									for(QDomNode n5 = n4.firstChild(); n5.isNull() == false; n5 = n5.nextSibling()) {

										// Gets the element (<usedInputs> | <usedOutputs> | <usedActions>)...
										e = n5.toElement();

										// If the element is the first occurrence of <usedInputs>...
										if(usedInputsFlag == false && e.isNull() == false && e.tagName() == "usedInputs") {

											// First occurrence of <usedInputs> element found
											usedInputsFlag = true;

											// For each <usedInputs> child (<usedInput>)
											QList<UsedInput *> *usedInputList = (QList<UsedInput *> *) usedInputs->getUsedInputList();
											for(QDomNode n6 = n5.firstChild(); n6.isNull() == false; n6 = n6.nextSibling()) {

												// <usedInput>
												e = n6.toElement();
												if(e.isNull() == false && e.tagName() == "usedInput") {

													// Sets the attributes
													UsedInput *usedInput = new UsedInput;
													usedInput->setInput(e.attribute("input", ""));
													usedInputList->append(usedInput);

												}

											}

										}

										// If the element is the first occurrence of <usedActions>...
										if(usedActionsFlag == false && e.isNull() == false && e.tagName() == "usedActions") {

											// First occurrence of <usedActions> element found
											usedActionsFlag = true;

											// For each <usedActions> child (<usedAction>)
											QList<UsedAction *> *usedActionList = (QList<UsedAction *> *) usedActions->getUsedActionList();
											for(QDomNode n6 = n5.firstChild(); n6.isNull() == false; n6 = n6.nextSibling()) {

												// <item>
												e = n6.toElement();
												if(e.isNull() == false && e.tagName() == "usedAction") {

													// Sets the attributes
													UsedAction *usedAction = new UsedAction;
													usedAction->setAction(e.attribute("action", ""));
													usedAction->setContainsRegex(e.attribute("containsRegex", ""));
													usedAction->setOnTrueOutput(e.attribute("onTrueOutput", ""));
													usedAction->setOnTrue(e.attribute("onTrue", ""));
													usedActionList->append(usedAction);

												}

											}

										}

									}

								}

							}

						}

					}

				}

			}

		}

	}

	// Returns the success
	return true;

}

const QString Kms::_pickGlobalText(const QString &string, const QString &language) {

	// Gets the default language
	const Texts *texts = _kmsRoot->getTexts();
	QString selectedLanguage;
	if(language != 0) {
		selectedLanguage = language;
	} else {
		selectedLanguage = texts->getDefaultLanguage();
	}

	// Finds the text into the text list
	const QList<Text *> *textList = texts->getTextList();
	int textListSize = textList->size();
	for(int i = 0; i < textListSize; i++) {
		const Text *text = textList->at(i);
		if(text->getId() == string) {
			// Finds the right translation of the text
			const QList<Value *> *valueList = text->getValueList();
			int valueListSize = valueList->size();
			for(int j = 0; j < valueListSize; j++) {
				const Value *value = valueList->at(j);
				if(value->getLanguage() == selectedLanguage) {
					return value->getValue();
				}
			}
		}
	}

	// Returns the result
	return "";

}

const QString Kms::_pickLocalText(const QString &serviceId, const QString &string, const QString &language) {

	// Gets the text list
	const QList<Service *> *serviceList = _kmsRoot->getServices()->getServiceList();
	int serviceListSize = serviceList->size();
	for(int i = 0; i < serviceListSize ; i++) {
		const Service *service = serviceList->at(i);
		if(service->getId() == serviceId) {
			const Texts *texts = service->getTexts();

			// Gets the default language
			QString selectedLanguage;
			if(language != 0) {
				selectedLanguage = language;
			} else {
				selectedLanguage = texts->getDefaultLanguage();
			}

			// Finds the text into the text list
			const QList<Text *> *textList = texts->getTextList();
			int textListSize = textList->size();
			for(int i = 0; i < textListSize; i++) {
				const Text *text = textList->at(i);
				if(text->getId() == string) {
					// Finds the right translation of the text
					const QList<Value *> *valueList = text->getValueList();
					int valueListSize = valueList->size();
					for(int j = 0; j < valueListSize; j++) {
						const Value *value = valueList->at(j);
						if(value->getLanguage() == selectedLanguage) {
							return value->getValue();
						}
					}
				}
			}

		}
	}

	// Returns the result
	return "";

}

const QString Kms::_pickText(const QString &serviceId, const QString &string, const QString &language) {

	// Evaluates the local text before and the global text later (if no local text has been found)
	QString result = _pickLocalText(serviceId, string, language);
	if(result == "") {
		result = _pickGlobalText(string, language);
	}

	// Returns the result
	return result;

}

bool Kms::_performSequence(const QString &serviceId, const QString &sequenceId, const QList< QPair<QString, QString> > &inputDataList) const {

	// This flag stops the execution in case of errors
	bool _failureFlag = false;

	// Searches the right service
	Service *currentService = 0;
	const QList<Service *> *serviceList = _kmsRoot->getServices()->getServiceList();
	int serviceListSize = serviceList->size();
	for(int i = 0; i < serviceListSize; i++) {
		Service *service = serviceList->at(i);
		if(service->getId() == serviceId) {
			currentService = service;
			break;
		}
	}

	// If the service has not been found
	if(currentService == 0) {
		qDebug("[kms::Kms] ERROR: Service not found!");
		return false;
	}

	// Searches the right sequence
	Sequence *currentSequence = 0;
	const QList<Sequence *> *sequenceList = currentService->getSequences()->getSequenceList();
	int sequenceListSize = sequenceList->size();
	for(int i = 0; i < sequenceListSize; i++) {
		Sequence *sequence = sequenceList->at(i);
		if(sequence->getId() == sequenceId) {
			currentSequence = sequence;
			break;
		}
	}

	// If the sequence has not been found
	if(currentSequence == 0) {
		qDebug("[kms::Kms] ERROR: Sequence not found!");
		return false;
	}

	// Gets the remaining inputs (if there are anyone)...
	QList< QHash<QString, QString> > remainingInputHashList;
	const QList<Input *> *inputList = getInputList(serviceId);
	const QList<UsedInput *> *usedInputList = getUsedInputList(serviceId, sequenceId);
	int inputListSize = inputList->size();
	int usedInputListSize = usedInputList->size();
	int inputDataListSize = inputDataList.size();
	for(int i = 0; i < inputListSize; i++) {
		Input *input = inputList->at(i);
		// Tests if this input is considered by the used inputs
		bool usedFlag = false;
		for(int j = 0; j < usedInputListSize; j++) {
			if(input->getId() == usedInputList->at(j)->getInput()) {
				usedFlag = true;
				break;
			}
		}
		if(usedFlag == false) {
			continue;
		}
		// ...
		bool probed = false;
		for(int j = 0; j < inputDataListSize; j++) {
			QPair<QString, QString> pair = inputDataList.at(j);
			if(pair.first == input->getId()) {
				probed = true;
				break;
			}
		}
		if(probed == false) {
			QHash<QString, QString> inputHash;
			inputHash.insert("id", input->getId());
			inputHash.insert("content", input->getContent());
			inputHash.insert("type", input->getType());
			inputHash.insert("regex", input->getRegex());
			QString name = ((Kms *) this)->evaluateText(serviceId, input->getName());
			inputHash.insert("name", name);
			inputHash.insert("defaultValue", input->getDefaultValue());
			remainingInputHashList.append(inputHash);
		}
	}

	// Asks the remaining inputs to the user
	bool ok;
	QList< QPair<QString, QString> > remainingInputDataList = _ioStrategy->getInputs(remainingInputHashList, &ok);
	if(ok == false) {
		return false;
	}

	// Gets the action list
	const QList<Action *> *currentActionList = currentService->getActions()->getActionList();
	int currentActionListSize = currentActionList->size();

	// For each used action...
	const QList<UsedAction *> *currentUsedActionList = currentSequence->getUsedActions()->getUsedActionList();
	int currentUsedActionListSize = currentUsedActionList->size();
	for(int i = 0; i < currentUsedActionListSize && _failureFlag == false; i++) {
		UsedAction *currentUsedAction = currentUsedActionList->at(i);
		for(int j = 0; j < currentActionListSize; j++) {
			Action *currentAction = currentActionList->at(j);
			if(currentUsedAction->getAction() == currentAction->getId()) {

				// Performs the single action
				if(currentAction->getType() == "httpRequest") {

					// Gets the http request as text
					HttpRequest *httpRequestTag = (HttpRequest *) currentAction->getHttpRequest();
					QString httpRequest = httpRequestTag->getText();

					// For each probed input data...
					for(int i = 0; i < inputDataList.size(); i++) {
						// Replaces all the occurences of the input id (pair.first) with its value (pair.second)
						QPair<QString, QString> pair = inputDataList.at(i);
						QString second = QUrl::toPercentEncoding(pair.second, QByteArray(" "));
						httpRequest.replace(QRegExp("\\$INPUT\\[(" + pair.first + ")\\]"), second);
					}

					// For each remaining input data...
					for(int i = 0; i < remainingInputDataList.size(); i++) {
						// Replaces all the occurences of the input id (pair.first) with its value (pair.second)
						QPair<QString, QString> pair = remainingInputDataList.at(i);
						QString second = QUrl::toPercentEncoding(pair.second, QByteArray(" "));
						httpRequest.replace(QRegExp("\\$INPUT\\[(" + pair.first + ")\\]"), second);
					}

					// For each contextual input...
					for(int i = 0; i < inputList->size(); i++) {
						Input *input = inputList->at(i);
						if(input->getType() == "contextual") {
							// Replaces all the occurences of the input id with its value
							httpRequest.replace(QRegExp("\\$INPUT\\[(" + input->getId() + ")\\]"), input->getTmpValue()); // TMP!
						}
					}

					// Stores the requestId
					((QString &) _currentHttpRequestId) = httpRequestTag->getId();

					// Performs the http request
					bool result = _doHttpRequest(httpRequest);
					if(result == false) {
						return false;
					}

					// Pauses the sequence
					_ioStrategy->showMessage("Click \"ok\" AFTER this action has completed...");

				}

				// Performs the single action
				if(currentAction->getType() == "checkHttpRequest") {

					// Gets the "checkHttpRequest"
					CheckHttpRequest *checkHttpRequest = (CheckHttpRequest *) currentAction->getCheckHttpRequest();
					QString httpRequest = checkHttpRequest->getHttpRequest();

					// Gets the data from the right item of the current http request data list
					QByteArray data;
					bool found = false;
					int size = _currentHttpRequestDataList.size();
					for(int i = 0; i < size; i++) {
						QPair<QString, QByteArray> item = _currentHttpRequestDataList.at(i);
						if(item.first == httpRequest) {
							data = item.second;
							found = true;
							break;
						}
					}

					// If the data exists
					if(found == true) {
						// Performs the check
						QString containsRegex = checkHttpRequest->getContainsRegex();
						QString onTrueOutput = checkHttpRequest->getOnTrueOutput();
						QString onFalseOutput = checkHttpRequest->getOnFalseOutput();
						QString onTrue = checkHttpRequest->getOnTrue();
						QString onFalse = checkHttpRequest->getOnFalse();
						if(containsRegex != "") {
							bool result = QRegExp(containsRegex).exactMatch(QString(data));
							if(result == true) {
								if(onTrueOutput != "") {
									_ioStrategy->showMessage(onTrueOutput);
								}
								if(onTrue != "") {
									if(onTrue == "continue") {
										// Nothing...
									}
									if(onTrue == "break") {
										_failureFlag = true;
										break;
									}
								}
							}
							if(result == false) {
								if(onFalseOutput != "") {
									_ioStrategy->showMessage(onFalseOutput);
								}
								if(onFalse != "") {
									if(onFalse == "continue") {
										// Nothing...
									}
									if(onFalse == "break") {
										_failureFlag = true;
										break;
									}
								}
							}
						}
					}

				}

				// Performs the single action
				if(currentAction->getType() == "askCaptchaCode") {

					// Gets the "askCaptchaCode"
					AskCaptchaCode *askCaptchaCode = (AskCaptchaCode *) currentAction->getAskCaptchaCode();
					QString httpRequest = askCaptchaCode->getHttpRequest();
					QString input = askCaptchaCode->getInput();

					// Gets the data from the right item of the current http request data list
					QByteArray data;
					bool found = false;
					int size = _currentHttpRequestDataList.size();
					for(int i = 0; i < size; i++) {
						QPair<QString, QByteArray> item = _currentHttpRequestDataList.at(i);
						if(item.first == httpRequest) {
							data = item.second;
							found = true;
							break;
						}
					}

					// If the data exists
					if(found == true) {
						const QList<Input *> *inputList = getInputList(serviceId);
						for(int i = 0; i < inputList->size(); i++) {
							Input *currentInput = inputList->at(i);
							if(currentInput->getId() == input) {
								QHash<QString, QString> hash;
								hash.insert("id", currentInput->getId());
								hash.insert("content", currentInput->getContent());
								hash.insert("type", currentInput->getType());
								hash.insert("regex", currentInput->getRegex());
								QString name = ((Kms *) this)->evaluateText(serviceId, currentInput->getName());
								hash.insert("name", name);
								hash.insert("defaultValue", currentInput->getDefaultValue());
								bool ok;
								QString result = _ioStrategy->getCaptchaInput(hash, data, &ok);
								if(ok == false) {
									_failureFlag = true;
									break;
								}
								currentInput->setTmpValue(result); // TMP!
							}
						}
					}

				}

			}
		}
	}

	// Returns the result
	return true;

}

bool Kms::_doHttpRequest(const QString &text) const {

	// Cerates the regular expression for the http request header
	QRegExp qRegExp("(\\s*)(GET|POST)(\\s)(.+)(\\s)(HTTP/\\d\\.\\d)(\\\\r\\\\n)(\\r?\\n)((\\s*)(.+:\\s.+)(\\\\r\\\\n)(\\r?\\n))*(\\s*)(\\\\r\\\\n)(\\r?\\n\\s*Data:\\s.*)?");

	// Gets the trimmed text
	QString request = text.trimmed();

	// Checks if it is a well-formed http request header
	bool result = qRegExp.exactMatch(request);
	if(result == false) {
		qDebug("[kms::Engine] ERROR: http request header is malformed!");
		return false;
	}

	// Splits the http request header in rows
	QStringList rows = request.split(QRegExp("\\r?\\n"));

	// Sets the method
	QString method;
	if(QRegExp("(\\s*)(GET)(\\s).*").exactMatch(rows.at(0)) == true) {
		method = "GET";
	} else {
		method = "POST";
	}

	// Sets the uri
	QString firstRow = rows.at(0);
	int uriHead = -1;
	if(method == "GET") {
		uriHead = firstRow.indexOf(QRegExp("GET")) + 4;
	} else {
		uriHead = firstRow.indexOf(QRegExp("POST")) + 5;
	}
	int uriTail = firstRow.indexOf(QRegExp("HTTP")) - 2;
	QString uri = firstRow.left(uriTail + 1);
	uri = uri.right(uri.size() - uriHead);

	// Creates and sets the http request header
	QHttpRequestHeader header(method, uri);

	// Prints a debug string
	cout << (method + " " + uri).toStdString() << endl;

	// Sets the data
	QString data;

	// Sets the values
	for(int i = 1; i < rows.size(); i++) {
		QString s = rows.at(i);
		if(QRegExp("(\\s*)(\\\\r\\\\n)").exactMatch(s) == true) {
			continue;
		}
		if(QRegExp("(\\s*)(Data:\\s)(.*)").exactMatch(s) == true) {
			s = s.trimmed();
			QStringList pair = s.split(QRegExp("Data:\\s"));
			data = pair[1];
			continue;
		}
		QStringList pair = s.split(QRegExp(":\\s"));
		pair[0] = pair[0].trimmed();
		pair[1] = pair[1].trimmed().left(pair[1].size() - 4);
		// Prints a debug string
		cout << pair[0].toStdString() << ": " << pair[1].toStdString() << endl;
		header.setValue(pair[0], pair[1]);
		if(pair[0] == "Host") {
			int index = pair[1].indexOf(QRegExp(":"));
			if(index == -1) {
				_enhancedQHttp->setHost(pair[1]);
			} else {
				_enhancedQHttp->setHost(pair[1].left(index), pair[1].right(pair[1].size() - (index + 1)).toInt());
			}
		}
	}

	// Attaches the cookies
	CookieManager *cookieManager = _enhancedQHttp->getCookieManager();
	cookieManager->attachCookies(&header);

	// Does the http request
	if(data == "") {
		_enhancedQHttp->request(header);
	} else {
		// Prints a debug string
		cout << QString("Data: ").toStdString() << data.toStdString() << endl;
		_enhancedQHttp->request(header, data.toUtf8());
	}

	// Returns the result
	return true;

}

void Kms::requestPerformed(QByteArray &data) {

	// Sets the data into the right item of the current http request data list
	int size = _currentHttpRequestDataList.size();
	for(int i = 0; i < size; i++) {
		QPair<QString, QByteArray> item = _currentHttpRequestDataList.at(i);
		if(item.first == _currentHttpRequestId) {
			QPair<QString, QByteArray> newPair(item.first, data);
			_currentHttpRequestDataList.replace(i, newPair);
			break;
		}
	}

}

/* OLD
void Kms::requestPerformed(QByteArray &data) {

	QString text = QString(data);
	cout << text.toStdString() << endl;

	QString successRegex = _currentHttpRequestDomElement.attribute("successRegex", "");
	QString outputOnSuccess = _currentHttpRequestDomElement.attribute("outputOnSuccess", "");
	if(successRegex != "" && QRegExp(successRegex).exactMatch(text) == true) {
		_ioStrategy->showMessage(outputOnSuccess);
	}

	QString failureRegex = _currentHttpRequestDomElement.attribute("failureRegex", "");
	QString outputOnFailure = _currentHttpRequestDomElement.attribute("outputOnFailure", "");
	if(failureRegex != "" && QRegExp(successRegex).exactMatch(text) == true) {
		_ioStrategy->showMessage(outputOnFailure);
		_failureFlag = true;
	}

	QString inputAttribute = _currentHttpRequestDomElement.attribute("input", "");
	if(inputAttribute != "") {
		QList<Input *> *inputList = getInputList(_currentServiceId, _currentSequenceId);
		for(int i = 0; i < inputList->size(); i++) {
			Input *input = inputList->at(i);
			if(input->getId() == inputAttribute) {
				QHash<QString, QString> hash;
				hash.insert(input->getId(), _currentHttpRequestDomElement.attribute("id", ""));
				hash.insert(input->getName(), _currentHttpRequestDomElement.attribute("name", ""));
				hash.insert(input->getType(), _currentHttpRequestDomElement.attribute("type", ""));
				hash.insert(input->getContent(), _currentHttpRequestDomElement.attribute("context", ""));
				hash.insert(input->getRegex(), _currentHttpRequestDomElement.attribute("regex", ""));
				hash.insert(input->getValue(), _currentHttpRequestDomElement.attribute("value", ""));
				hash.insert(input->getValues(), _currentHttpRequestDomElement.attribute("values", ""));
				QString result = _ioStrategy->getCaptchaInput(hash, data);
				input->setValue(result);
			}
		}
	}

}
*/
