/*
 * kms.h
 *
 * @author Pietro Spinella (scrockman[at]gmail[dot]com)
 * @license GPLv3
 */

#ifndef KMS_KMS_H
#define KMS_KMS_H

#include <QDomDocument>
#include <QHttpRequestHeader>
#include <QList>

#include "enhancedqhttp/enhancedqhttp.h"
#include "iostrategy.h"
#include "tree/kmsroot.h"

namespace kms {

	class Kms : public QObject {

		Q_OBJECT

		public:
			Kms();
			~Kms();

			// Sets the input/output strategy (see strategy pattern on google for more details)
			void setIOStrategy(const IOStrategy *ioStrategy);
			// Tryes to load the specified xml file if and only if the engine is not running
			bool load(const QString &fileName);

			// Gets the input list of the specified service
			const QList<Input *> *getInputList(const QString &serviceId) const;
			// Gets the used input list of the specified sequence (of the specified service)
			const QList<UsedInput *> *getUsedInputList(const QString &serviceId, const QString &sequenceId) const;

			// Gets the input count (of a specified type) of the specified service
			int getInputCount(const QString &serviceId, const QString &type = 0) const;
			// Gets the used input count (of a specified type) of the specified sequence (of a specified service)
			int getUsedInputCount(const QString &serviceId, const QString &sequenceId, const QString &type = 0) const;

			// Evaluates the specified text (string) using the texts of the specified service (serviceId) translating it into the specified language (language)
			const QString evaluateText(const QString &serviceId, const QString &string, const QString &language = 0);

			// Performs the specified sequence (of a specified service), the inputs will be asked to the user
			bool performSequence(const QString &serviceId, const QString &sequenceId) const;
			// Performs the specified sequence (of a specified service) probing all (or some of) the inputs, the remaining inputs will be asked to the user
			bool performSequence(const QString &serviceId, const QString &sequenceId, const QList< QPair<QString, QString> > &inputDataList) const;

		private:
			// Extends the QHttp allowing redirects and cookie management
			EnhancedQHttp *_enhancedQHttp;

			// Stores the input/output strategy
			IOStrategy *_ioStrategy;

			// Stores an input list and an used input list used as null objects
			QList<Input *> *_inputListNullObject;
			QList<UsedInput *> *_usedInputListNullObject;

			// Stores the tree
			KmsRoot *_kmsRoot;

			// Stores the engine running status
			bool _runningFlag;

			// Used by load(.) to fill the tree
			bool _fillTree(const QDomDocument *qDomDocument);

			// Used by evaluateText(.) to pick the right text from the tree
			const QString _pickGlobalText(const QString &string, const QString &language = 0);
			const QString _pickLocalText(const QString &serviceId, const QString &string, const QString &language = 0);
			const QString _pickText(const QString &serviceId, const QString &string, const QString &language = 0);

			// Does the real performSequence method
			bool _performSequence(const QString &serviceId, const QString &sequenceId, const QList< QPair<QString, QString> > &inputDataList) const;

			// Performs the specified (as text) http request
			bool _doHttpRequest(const QString &text) const;

		// From here all the fields are related with the last invocation of the performSequence method...

			// Stores the current http request id
			QString _currentHttpRequestId;

			// Stores the (httpRequestId, httpRequestData) pairs
			QList< QPair<QString, QByteArray> > _currentHttpRequestDataList;

		signals:
			void sequenceFinished();

		private slots:
			void requestPerformed(QByteArray &data);

	};

};

#endif
