/*
 * mainwindow.cpp
 *
 * @author Pietro Spinella (scrockman[at]gmail[dot]com)
 * @license GPLv3
 */

#include <QDesktopServices>
#include <QDir>
#include <QDomDocument>
#include <QDomNode>
#include <QDomNodeList>
#include <QFile>
#include <QIODevice>
#include <QPalette>
#include <QTableWidgetItem>
#include <QTextStream>
#include <QUrl>

#include "mainwindow.h"
#include "accountmanager.h"

#include "telephonenumber.h"
#include "accountitemdelegate.h"
#include "indexbookitemdelegate.h"
#include "scheduleritemdelegate.h"

#include "providers/aliceitsms.h"
#include "providers/tinitsms.h"
#include "providers/vodafoneitmms.h"
#include "providers/vodafoneitsms.h"

using namespace kms;

kms::MainWindow::MainWindow() : QMainWindow() {

	_accountManager = new AccountManager;	// To delete
//	_accountManager->setVisible(true);

	// Sets the user interface
	_ui.setupUi(this);

	// Creates the captcha manager
	_captchaManager = new CaptchaManager;

	// Creates the debug text edit
	debugTextEdit = new QTextEdit;
	debugTextEdit->resize(400, 300);
//	debugTextEdit->adjustSize();
//	debugTextEdit->setAcceptRichText(false);

	// Creates and starts the scheduler
	_scheduler = new Scheduler(this);
	_scheduler->start();

	// Creates and sets the tray icon
	if(QSystemTrayIcon::isSystemTrayAvailable() == true) {

		// Create and set the quit action for the tray icon menu
		_quitAction = new QAction(QApplication::translate("MainWindow", "&Quit", 0, QApplication::UnicodeUTF8), this);
		_quitAction->setIcon(QIcon(QString::fromUtf8(":/images/images/error.png")));
		connect(_quitAction, SIGNAL(triggered()), this, SLOT(slot_saveAllAndQuit()));

		// Create and set the menu for the tray icon
		_trayIconMenu = new QMenu(this);
		_trayIconMenu->addSeparator();
		_trayIconMenu->addAction(_quitAction);

		// Create, set and show the tray icon
		_trayIcon = new QSystemTrayIcon(this);
		_trayIcon->setIcon(QIcon(QString::fromUtf8(":/images/images/send.png")));
		_trayIcon->setContextMenu(_trayIconMenu);
		_trayIcon->show();

		QApplication::connect(_trayIcon, SIGNAL(activated(QSystemTrayIcon::ActivationReason)), this, SLOT(slot_iconActivated(QSystemTrayIcon::ActivationReason)));

	}

	// Create the dictionary for the packing algorithms
	_dictionary.append(QPair<QString, QString>("per","x"));
	_dictionary.append(QPair<QString, QString>("Per","X"));
	_dictionary.append(QPair<QString, QString>("ch","k"));
	_dictionary.append(QPair<QString, QString>("Ch","K"));
	_dictionary.append(QPair<QString, QString>("con","kn"));
	_dictionary.append(QPair<QString, QString>("con","Kn"));
	_dictionary.append(QPair<QString, QString>("solo","sl"));
	_dictionary.append(QPair<QString, QString>("Solo","Sl"));
	_dictionary.append(QPair<QString, QString>("piu'","+"));
	_dictionary.append(QPair<QString, QString>("Piu'","+"));
	_dictionary.append(QPair<QString, QString>("piu","+"));
	_dictionary.append(QPair<QString, QString>("Piu","+"));
	_dictionary.append(QPair<QString, QString>("meno","-"));
	_dictionary.append(QPair<QString, QString>("Meno","-"));
	_dictionary.append(QPair<QString, QString>("una","1"));
	_dictionary.append(QPair<QString, QString>("Una","1"));
	_dictionary.append(QPair<QString, QString>("uno","1"));
	_dictionary.append(QPair<QString, QString>("Uno","1"));
	_dictionary.append(QPair<QString, QString>("un","1"));
	_dictionary.append(QPair<QString, QString>("Un","1"));
	_dictionary.append(QPair<QString, QString>("due","2"));
	_dictionary.append(QPair<QString, QString>("Due","2"));
	_dictionary.append(QPair<QString, QString>("tre","3"));
	_dictionary.append(QPair<QString, QString>("Tre","3"));
	_dictionary.append(QPair<QString, QString>("quattro","4"));
	_dictionary.append(QPair<QString, QString>("Quattro","4"));
	_dictionary.append(QPair<QString, QString>("cinque","5"));
	_dictionary.append(QPair<QString, QString>("Cinque","5"));
	_dictionary.append(QPair<QString, QString>("sei","6"));
	_dictionary.append(QPair<QString, QString>("Sei","6"));
	_dictionary.append(QPair<QString, QString>("sette","7"));
	_dictionary.append(QPair<QString, QString>("Sette","7"));
	_dictionary.append(QPair<QString, QString>("otto","8"));
	_dictionary.append(QPair<QString, QString>("Otto","8"));
	_dictionary.append(QPair<QString, QString>("nove","9"));
	_dictionary.append(QPair<QString, QString>("Nove","9"));
	_dictionary.append(QPair<QString, QString>("dieci","10"));
	_dictionary.append(QPair<QString, QString>("Dieci","10"));
	_dictionary.append(QPair<QString, QString>("punto","\".\""));
	_dictionary.append(QPair<QString, QString>("Punto","\".\""));

	// Set the dictionary for the packing algorithms
	_simpleAlgorithm.setDictionary(&_dictionary);
	_compactAlgorithm.setDictionary(&_dictionary);
	_crypticAlgorithm.setDictionary(&_dictionary);
	_nastyAlgorithm.setDictionary(&_dictionary);

	// Sets the default packing algorithm for the text packer
//c	_textPacker.giveDefaultPackingAlgorithm(&_nastyAlgorithm);	// Take the ownership

	// Set some variables for the text packing feature
	_lastPackingIndex = 0;

	// Create and set the index book table widget
	QStringList indexBookTableWidgetLabels;
	indexBookTableWidgetLabels << QApplication::translate("MainWindow", "Name", 0, QApplication::UnicodeUTF8) <<
					QApplication::translate("MainWindow", "Number", 0, QApplication::UnicodeUTF8) <<
					QApplication::translate("MainWindow", "Predefined account", 0, QApplication::UnicodeUTF8);
	_ui.indexBookTableWidget->setRowCount(0);
	_ui.indexBookTableWidget->setColumnCount(3);
	_ui.indexBookTableWidget->horizontalHeader()->resizeSection(0, 200);
	_ui.indexBookTableWidget->horizontalHeader()->resizeSection(1, 200);
	_ui.indexBookTableWidget->horizontalHeader()->resizeSection(2, 200);
	_ui.indexBookTableWidget->setSelectionMode(QAbstractItemView::NoSelection);
	_ui.indexBookTableWidget->setItemDelegate(new IndexBookItemDelegate(this));	//Non liberato
	_ui.indexBookTableWidget->setHorizontalHeaderLabels(indexBookTableWidgetLabels);
	_ui.indexBookTableWidget->verticalHeader()->hide();
	_ui.indexBookTableWidget->sortByColumn(0, Qt::AscendingOrder);

	// Create and set the account table widget
	QStringList accountQTableWidgetLabels;
	accountQTableWidgetLabels << QApplication::translate("MainWindow", "Name", 0, QApplication::UnicodeUTF8) <<
					QApplication::translate("MainWindow", "Type", 0, QApplication::UnicodeUTF8) <<
					QApplication::translate("MainWindow", "Username", 0, QApplication::UnicodeUTF8) <<
					QApplication::translate("MainWindow", "Password", 0, QApplication::UnicodeUTF8);
	_ui.accountQTableWidget->setRowCount(0);
	_ui.accountQTableWidget->setColumnCount(4);
	_ui.accountQTableWidget->horizontalHeader()->resizeSection(0, 200);
	_ui.accountQTableWidget->horizontalHeader()->resizeSection(1, 200);
	_ui.accountQTableWidget->horizontalHeader()->resizeSection(2, 200);
	_ui.accountQTableWidget->horizontalHeader()->resizeSection(3, 200);
	_ui.accountQTableWidget->setSelectionMode(QAbstractItemView::NoSelection);
	_ui.accountQTableWidget->setHorizontalHeaderLabels(accountQTableWidgetLabels);
	_ui.accountQTableWidget->verticalHeader()->hide();
	// Create and set the account item delegate
	_accountItemDelegate = new AccountItemDelegate(this);
	_ui.accountQTableWidget->setItemDelegate(_accountItemDelegate);

	// Creates and sets the scheduler tree view
	QStandardItemModel *model = new QStandardItemModel(0, 5, this);
	model->setHeaderData(0, Qt::Horizontal, QObject::tr("Date"));
	model->setHeaderData(1, Qt::Horizontal, QObject::tr("Account"));
	model->setHeaderData(2, Qt::Horizontal, QObject::tr("Recipient"));
	model->setHeaderData(3, Qt::Horizontal, QObject::tr("Message"));
	model->setHeaderData(4, Qt::Horizontal, QObject::tr("Status"));
	_ui.schedulerQTreeView->setRootIsDecorated(false);
	_ui.schedulerQTreeView->setAlternatingRowColors(true);
	_ui.schedulerQTreeView->setModel(model);

	// ...
	_ui.recipientTelephoneNumber->setCountry(TelephoneNumber::Italy);
	_recipientQLineEdit = _ui.recipientTelephoneNumber->getTelephoneNumberQLineEdit();

	// Create the .scrocksms directory if it doesn't exists
	QDir dir = QDir::home();
	if(dir.exists(".scrocksms") == false) {
		dir.mkdir(".scrocksms");
	}

	// Load user data
	loadSettings();
	loadIndexBook();
	loadModels();

	// Connect the signals to the slots
//	connect(qApp, SIGNAL(lastWindowClosed()), this, SLOT(hide()/*slot_saveAllAndQuit()*/));
	qApp->setQuitOnLastWindowClosed(false);

	connect(_ui.webSiteQPushButton, SIGNAL(clicked()), this, SLOT(slot_webSiteQPushButtonClicked()));
	connect(_ui.forumQPushButton, SIGNAL(clicked()), this, SLOT(slot_forumQPushButtonClicked()));
	connect(_ui.accountQComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(slot_accountQComboBoxCurrentIndexChanged(int)));
	connect(_ui.packingQComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(slot_packingQComboBoxCurrentIndexChanged(int)));
	connect(_ui.recipientTelephoneNumber->getTelephoneNumberQLineEdit(), SIGNAL(textChanged(QString)), this, SLOT(slot_recipientQLineEditTextChanged(QString)));
	connect(_ui.senderQLineEdit, SIGNAL(textChanged(QString)), this, SLOT(slot_senderQLineEditTextChanged(QString)));
	connect(_ui.textQTextEdit, SIGNAL(textChanged()), this, SLOT(slot_textQTextEditChanged()));
	connect(_ui.sendQPushButton, SIGNAL(clicked()), this, SLOT(slot_sendQPushButtonClicked()));

	// Indexbook
	connect(_ui.useAsReceiverQPushButton, SIGNAL(clicked()), this, SLOT(slot_useAsReceiverQPushButtonClicked()));
	connect(_ui.newQPushButton, SIGNAL(clicked()), this, SLOT(slot_newQPushButtonClicked()));
	connect(_ui.removeQPushButton, SIGNAL(clicked()), this, SLOT(slot_removeQPushButtonClicked()));
	connect(_ui.indexBookTableWidget, SIGNAL(cellChanged(int, int)), this, SLOT(slot_indexBookTableWidgetCellChanged(int, int)));

	// Models
	connect(_ui.useAsTextQPushButton, SIGNAL(clicked()), this, SLOT(slot_useAsTextQPushButtonClicked()));
	connect(_ui.newModelQPushButton, SIGNAL(clicked()), this, SLOT(slot_newModelQPushButtonClicked()));
	connect(_ui.removeModelQPushButton, SIGNAL(clicked()), this, SLOT(slot_removeModelQPushButtonClicked()));
	connect(_ui.modelsQListWidget, SIGNAL(itemDoubleClicked(QListWidgetItem *)), this, SLOT(slot_modelsQListWidgetItemDoubleClicked(QListWidgetItem *)));
	connect(_ui.modelsQListWidget, SIGNAL(itemSelectionChanged()), this, SLOT(slot_modelsQListWidgetItemSelectionChanged()));

// TMP	connect(_ui.recipientTelephoneNumber->getTelephoneNumberQLineEdit(), SIGNAL(textChanged(QString)), this, SLOT(_updateStatus()));
// TMP	connect(_ui.showAccountInfoQCheckBox, SIGNAL(stateChanged(int)), this, SLOT(slot_showAccountInfoStateChanged(int)));

// Da qua ok...

	// Set the provider list
	_providerList.append(new AliceItSms);
	_providerList.append(new TinItSms);
	_providerList.append(new VodafoneItMms);
	_providerList.append(new VodafoneItSms);

	// Set the account list
	_accountList.append(new Account("Your - www.alice.it SMS", "", "", _providerList[0]));
	_accountList.append(new Account("Your - www.tin.it SMS", "", "", _providerList[1]));
	_accountList.append(new Account("Generic - www.vodafone.it MMS", "", "", _providerList[2]));
	_accountList.append(new Account("Your - www.vodafone.it SMS", "", "", _providerList[3]));

	// Set the account list into the account table widget
	for(QList<Account *>::iterator i = _accountList.begin(); i != _accountList.end(); i++) {
		int row = _ui.accountQTableWidget->rowCount();
		_ui.accountQTableWidget->setRowCount(row + 1);
		QTableWidgetItem *item0 = new QTableWidgetItem((*i)->getName());
		QTableWidgetItem *item1 = new QTableWidgetItem((*i)->getProvider()->getName());
		QTableWidgetItem *item2 = new QTableWidgetItem((*i)->getUsername());
		QTableWidgetItem *item3 = new QTableWidgetItem((*i)->getPassword());
		_ui.accountQTableWidget->setItem(row, 0, item0);
		_ui.accountQTableWidget->setItem(row, 1, item1);
		_ui.accountQTableWidget->setItem(row, 2, item2);
		_ui.accountQTableWidget->setItem(row, 3, item3);
		_ui.accountQTableWidget->openPersistentEditor(item1);
	}

	// Set the account list into the combo box
	for(QList<Account *>::iterator i = _accountList.begin(); i != _accountList.end(); i++) {
		_ui.accountQComboBox->addItem(QIcon(QString::fromUtf8(":/images/images/account.png")), (*i)->getName());
	}

	// Set the focus on the recipient line edit
	_recipientQLineEdit->setFocus();

}

kms::MainWindow::~MainWindow() {

	// Destroy the tray icon and all its components
	if(QSystemTrayIcon::isSystemTrayAvailable() == true) {
		delete _trayIcon;
		delete _trayIconMenu;
		disconnect(_quitAction, 0, 0, 0);
		delete _quitAction;
	}

	// Destroy the account list
	for(QList<Account *>::iterator i = _accountList.begin(); i != _accountList.end(); i++) {
		delete *i;
	}

	// Destroy the provider list
	for(QList<Provider *>::iterator i = _providerList.begin(); i != _providerList.end(); i++) {
		delete *i;
	}

	// Destroy the account item delegate
	delete _accountItemDelegate;

	// Destroyes the scheduler (the scheduler destroyes itself)
	_scheduler->isAlive = false;

	// Destroyes the captcha manager;
	delete _captchaManager;

}

CaptchaManager *kms::MainWindow::getCaptchaManager() {
	return _captchaManager;
}

void kms::MainWindow::loadSettings() {

	QDomDocument settingsQDomDocument("settings");
	QFile file(QDir::homePath() + "/.scrocksms/settings.xml");
	if(file.exists()==true) {
		if(file.open(QIODevice::ReadOnly)==true) {
			if(settingsQDomDocument.setContent(&file)==true) {
				QDomNodeList qDomNodeList;
				// General
				qDomNodeList = settingsQDomDocument.elementsByTagName("general").item(0).childNodes();
				for(int i = 0; i < qDomNodeList.count(); i++) {
					QDomNode qDomNode = qDomNodeList.item(i);
					QString value = qDomNode.firstChild().nodeValue().trimmed();
					if(qDomNode.nodeName()=="checksupdates") {
						value == "true" ? _ui.checksUpdatesQCheckBox->setCheckState(Qt::Checked) : _ui.checksUpdatesQCheckBox->setCheckState(Qt::Unchecked);
					}
					if(qDomNode.nodeName()=="showaccountinfo") {
						value == "true" ? _ui.showAccountInfoQCheckBox->setCheckState(Qt::Checked) : _ui.showAccountInfoQCheckBox->setCheckState(Qt::Unchecked);
					}
				}
				// Accounts
				qDomNodeList = settingsQDomDocument.elementsByTagName("account").item(0).childNodes();
				for(int i = 0; i < qDomNodeList.count(); i++) {
					QDomNode qDomNode = qDomNodeList.item(i);
					QString value = qDomNode.firstChild().nodeValue().trimmed();
					if(qDomNode.nodeName()=="username") {
//tmp						_ui.usernameQLineEdit->setText(value);
					}
					if(qDomNode.nodeName()=="password") {
//tmp						_ui.passwordQLineEdit->setText(value);
					}
				}
			}
			file.close();
		}
	}

}

void kms::MainWindow::saveSettings() {

	QDomDocument settingsQDomDocument("settings");
	QFile file(QDir::homePath() + "/.scrocksms/settings.xml");
	if(file.open(QIODevice::WriteOnly)==true) {
		QDomElement rootTag = settingsQDomDocument.createElement("root");
		QDomElement settingsTag = settingsQDomDocument.createElement("settings");
		QDomElement generalTag = settingsQDomDocument.createElement("general");
		QDomElement languageTag = settingsQDomDocument.createElement("language");
		QDomElement checksUpdatesTag = settingsQDomDocument.createElement("checksupdates");
		QDomElement showAccountInfoTag = settingsQDomDocument.createElement("showaccountinfo");
		QDomElement accountsTag = settingsQDomDocument.createElement("accounts");
		QDomElement accountTag = settingsQDomDocument.createElement("account");
		QDomElement usernameTag = settingsQDomDocument.createElement("username");
		QDomElement passwordTag = settingsQDomDocument.createElement("password");
		QDomText checksUpdatesValue = settingsQDomDocument.createTextNode((_ui.checksUpdatesQCheckBox->checkState()==Qt::Checked) ? "true" : "false");
		QDomText showAccountInfoValue = settingsQDomDocument.createTextNode((_ui.showAccountInfoQCheckBox->checkState()==Qt::Checked) ? "true" : "false");
		QDomText usernameValue = settingsQDomDocument.createTextNode(""/*_ui.usernameQLineEdit->displayText()*/);
		QDomText passwordValue = settingsQDomDocument.createTextNode(""/*_ui.passwordQLineEdit->text()*/);
		settingsQDomDocument.appendChild(rootTag);
			rootTag.appendChild(settingsTag);
				settingsTag.appendChild(generalTag);
					generalTag.appendChild(checksUpdatesTag);
						checksUpdatesTag.appendChild(checksUpdatesValue);
					generalTag.appendChild(showAccountInfoTag);
						showAccountInfoTag.appendChild(showAccountInfoValue);
				settingsTag.appendChild(accountsTag);
					accountsTag.appendChild(accountTag);
						accountTag.appendChild(usernameTag);
							usernameTag.appendChild(usernameValue);
							accountTag.appendChild(passwordTag);
							passwordTag.appendChild(passwordValue);
		QTextStream qTextStream(&file);
		settingsQDomDocument.save(qTextStream,2);
		file.close();
	}

}

void kms::MainWindow::loadIndexBook() {

	QDomDocument settingsQDomDocument("indexbook");
	QFile file(QDir::homePath() + "/.scrocksms/indexbook.xml");
	if(file.exists()==true) {
		if(file.open(QIODevice::ReadOnly)==true) {
			if(settingsQDomDocument.setContent(&file)==true) {
				// Contacts
				QDomNodeList qDomNodeList;
				QDomNodeList qDomNodeList1 = settingsQDomDocument.elementsByTagName("contacts").item(0).childNodes();
				for(int j = 0; j < qDomNodeList1.count(); j++) {
					qDomNodeList = qDomNodeList1.at(j).childNodes();
					QString name;
					QString telephoneNumber;
					QString predefinedAccount;
					for(int i = 0; i < qDomNodeList.count(); i++) {
						QDomNode qDomNode = qDomNodeList.item(i);
						QString value = qDomNode.firstChild().nodeValue().trimmed();
						if(qDomNode.nodeName()=="name") {
							name = value;
							continue;
						}
						if(qDomNode.nodeName()=="telephonenumber") {
							telephoneNumber = value;
							continue;
						}
						if(qDomNode.nodeName()=="predefinedaccount") {
							predefinedAccount = value;
							continue;
						}
					}
					if(name != "" && telephoneNumber != "") {
						int row = _ui.indexBookTableWidget->rowCount();
						_ui.indexBookTableWidget->setRowCount(row + 1);
						QTableWidgetItem *item0 = new QTableWidgetItem(name);
						QTableWidgetItem *item1 = new QTableWidgetItem(telephoneNumber);
						QTableWidgetItem *item2 = new QTableWidgetItem(predefinedAccount);
						item0->setCheckState(Qt::Unchecked);
						_ui.indexBookTableWidget->setItem(row, 2, item2);
						_ui.indexBookTableWidget->setItem(row, 1, item1);
						_ui.indexBookTableWidget->setItem(row, 0, item0);
						_ui.indexBookTableWidget->openPersistentEditor(item2);
					}
				}
			}
			file.close();
		}
	}

}

void kms::MainWindow::saveIndexBook() {

	QDomDocument indexBookQDomDocument("indexbook");
	QFile file(QDir::homePath() + "/.scrocksms/indexbook.xml");
	if(file.open(QIODevice::WriteOnly)==true) {
		QDomElement root = indexBookQDomDocument.createElement("root");
		QDomElement indexBookTag = indexBookQDomDocument.createElement("indexbook");
		QDomElement contactsTag = indexBookQDomDocument.createElement("contacts");
		QDomElement contactTag = indexBookQDomDocument.createElement("contact");
		QDomElement nameTag = indexBookQDomDocument.createElement("name");
		QDomElement telephoneNumberTag = indexBookQDomDocument.createElement("telephonenumber");
		QDomElement accountTag = indexBookQDomDocument.createElement("predefinedaccount");
		indexBookQDomDocument.appendChild(root);
			root.appendChild(indexBookTag);
				indexBookTag.appendChild(contactsTag);
					QAbstractItemModel *model = _ui.indexBookTableWidget->model();
					int rowCount = _ui.indexBookTableWidget->rowCount();
					for(int i = 0; i < rowCount; i++) {
						QDomNode currentContactTag = contactTag.cloneNode();
						contactsTag.appendChild(currentContactTag);
							QDomNode currentNameTag = nameTag.cloneNode();
							currentContactTag.appendChild(currentNameTag);
								QDomText nameValue = indexBookQDomDocument.createTextNode(model->data(model->index(i, 0)).toString());
								currentNameTag.appendChild(nameValue);
							QDomNode currentTelephoneNumberTag = telephoneNumberTag.cloneNode();
							currentContactTag.appendChild(currentTelephoneNumberTag);
								QDomText telephoneNumberValue = indexBookQDomDocument.createTextNode(model->data(model->index(i, 1)).toString());
								currentTelephoneNumberTag.appendChild(telephoneNumberValue);
							QDomNode currentAccountTag = accountTag.cloneNode();
							currentContactTag.appendChild(currentAccountTag);
								QDomText accountValue = indexBookQDomDocument.createTextNode(model->data(model->index(i, 2)).toString());
								currentAccountTag.appendChild(accountValue);
					}
			QTextStream qTextStream(&file);
			indexBookQDomDocument.save(qTextStream,2);
		file.close();
	}

}

void kms::MainWindow::loadModels() {

	QDomDocument modelsQDomDocument("models");
	QFile file(QDir::homePath() + "/.scrocksms/models.xml");
	if(file.exists()==true) {
		if(file.open(QIODevice::ReadOnly)==true) {
			if(modelsQDomDocument.setContent(&file)==true) {
				// Models
				QDomNodeList qDomNodeList;
				QDomNodeList qDomNodeList1 = modelsQDomDocument.elementsByTagName("models").item(0).childNodes();
				for(int j = 0; j < qDomNodeList1.count(); j++) {
					qDomNodeList = qDomNodeList1.at(j).childNodes();
					QString text;
					for(int i = 0; i < qDomNodeList.count(); i++) {
						QDomNode qDomNode = qDomNodeList.item(i);
						QString value = qDomNode.firstChild().nodeValue().trimmed();
						if(qDomNode.nodeName()=="text") {
							text = value;
							continue;
						}
					}
					if(text != "") {
						_ui.modelsQListWidget->addItem(text);
					}
				}
			}
			file.close();
		}
	}

}

void kms::MainWindow::saveModels() {

	QDomDocument modelsQDomDocument("models");
	QFile file(QDir::homePath() + "/.scrocksms/models.xml");
	if(file.open(QIODevice::WriteOnly)==true) {
		QDomElement root = modelsQDomDocument.createElement("root");
		QDomElement modelsTag = modelsQDomDocument.createElement("models");
		QDomElement modelTag = modelsQDomDocument.createElement("model");
		QDomElement textTag = modelsQDomDocument.createElement("text");
		modelsQDomDocument.appendChild(root);
			root.appendChild(modelsTag);
				int count = _ui.modelsQListWidget->count();
				for(int i = 0; i < count; i++) {
					QDomNode currentModelTag = modelTag.cloneNode();
					modelsTag.appendChild(currentModelTag);
						QDomNode currentTextTag = textTag.cloneNode();
						currentModelTag.appendChild(currentTextTag);
							QDomText textValue = modelsQDomDocument.createTextNode(_ui.modelsQListWidget->item(i)->text());
							currentTextTag.appendChild(textValue);
				}
			QTextStream qTextStream(&file);
			modelsQDomDocument.save(qTextStream, 2);
		file.close();
	}

}

void kms::MainWindow::_updateStatus() {

	int currentUsedChars = _ui.textQTextEdit->toPlainText().size();
	int currentMaxSmsLen = _currentAccount->getProvider()->getMaxTextLenght();

	// Change QTextEdit color
/*	QPalette qPalette = _ui.textQTextEdit->palette();
	if(size <= _currentMaxSmsLen) {
		qPalette.setColor(QPalette::Base,QColor(255,255,255));
		qPalette.setColor(QPalette::Text,QColor(0,0,0));
//tmp		_ui.sendQPushButton->setEnabled(true);
	} else {
		qPalette.setColor(QPalette::Base,QColor(255,0,0));
		qPalette.setColor(QPalette::Text,QColor(255,255,255));
//tmp		_ui.sendQPushButton->setEnabled(false);
	}
	_ui.textQTextEdit->setPalette(qPalette);
*/
	// Change QStatusBar text
	QString qString("");
	if(currentMaxSmsLen != -1) {
		qString.sprintf("Used characters: %d/%d", currentUsedChars, currentMaxSmsLen);
	} else {
		qString.sprintf("Used characters: %d", currentUsedChars);

	}
	_ui.usedCharactersQLabel->setText(qString);
//old	_ui.qStatusBar->showMessage(qString);

	// Set the send button status
	Provider *provider = _currentAccount->getProvider();
	bool result = true;
	// Sender
	if(provider->acceptsSender() == true) {
		QString sender = _ui.senderQLineEdit->text();
		if((sender.size() > provider->getMaxSenderLenght() && provider->getMaxSenderLenght() != -1) ||
							provider->getSenderRegExp().exactMatch(sender) == false) {
			result = false;
		}
	}
	// Recipients
	QString recipient = _ui.recipientTelephoneNumber->getTelephoneNumberQLineEdit()->text();
	QStringList recipientList = recipient.split(QRegExp(" *, *"));
	for(int i = 0; i < recipientList.size(); i++) {
		QString currentRecipient = recipientList[i].trimmed();
		if(provider->acceptsRecipient() == true) {
			QString recipient = _ui.recipientTelephoneNumber->getTelephoneNumberQLineEdit()->text();
			if((recipient.size() > provider->getMaxRecipientLenght() && provider->getMaxRecipientLenght() != -1) ||
								provider->getRecipientRegExp().exactMatch(currentRecipient) == false) {
				result = false;
			}
		}
	}
	// Subject
	if(provider->acceptsSubject() == true) {
		QString subject = _ui.subjectQLineEdit->text();
		if((subject.size() > provider->getMaxSubjectLenght() && provider->getMaxSubjectLenght() != -1) ||
							provider->getSubjectRegExp().exactMatch(subject) == false) {
			result = false;
		}
	}
	// Text
	if(provider->acceptsText() == true) {
		QString text = _ui.textQTextEdit->toPlainText();
		if((text.size() > provider->getMaxTextLenght() && provider->getMaxTextLenght() != -1) ||
							provider->getTextRegExp().exactMatch(text) == false) {
			result = false;
		}
	}
	// Set the send button status
	_ui.sendQPushButton->setEnabled(result);

}

void kms::MainWindow::slot_saveAllAndQuit() {

	// Save user data
	saveSettings();
	saveIndexBook();
	saveModels();

	// Exit from the application
	QApplication::exit(0);

}

void kms::MainWindow::slot_iconActivated(QSystemTrayIcon::ActivationReason reason) {

	// CAMBIARE POSTO
	// Set the tray icon menu actions
	if(QSystemTrayIcon::isSystemTrayAvailable() == true) {

		_trayIconMenu->clear();

		// TST
		QAbstractItemModel *model = _ui.indexBookTableWidget->model();
		int rowCount = _ui.indexBookTableWidget->rowCount();
		for(int i = 0; i < rowCount; i++) {
			QAction *action = new QAction(model->data(model->index(i, 0)).toString(), this);
			action->setIcon(QIcon(QString::fromUtf8(":/images/images/username.png")));
//NO!			action->activate(QAction::Trigger);
			connect(action, SIGNAL(triggered()), this, SLOT(slot_trayIconActionPerformed()));
			_trayIconMenu->addAction(action);


		}

		// Create and set the separator
		_trayIconMenu->addSeparator();

		// Create and set the quit action
		_trayIconMenu->addAction(_quitAction);

	}

	// Manage the tray icon events
	switch(reason) {
		case QSystemTrayIcon::Trigger:
			this->setVisible(!this->isVisible());
			break;
		default:
			break;
	}

}

// DA TOGLIERE
void kms::MainWindow::slot_trayIconActionPerformed() {

//	QString string = _trayIconMenu->activeAction()->text();
//	QString string = "Pietro [vodafone]";
//	if(string == "Pietro [vodafone]") {
//		std::cout << "Pietro [vodafone]" << std::endl;
//	}

}

void kms::MainWindow::slot_webSiteQPushButtonClicked() {

	// Go to the web site
	QDesktopServices::openUrl(QUrl("http://scrockman.altervista.org/products/scrocksms/index.php"));

}

void kms::MainWindow::slot_forumQPushButtonClicked() {

	// Go to the forum
	QDesktopServices::openUrl(QUrl("http://scrockman.altervista.org/forum/index.php"));

}

void kms::MainWindow::slot_recipientQLineEditTextChanged(QString /*text*/) {

	_updateStatus();

}

void kms::MainWindow::slot_senderQLineEditTextChanged(QString /*text*/) {

	_updateStatus();

}

void kms::MainWindow::slot_textQTextEditChanged() {

	if(_ui.textQTextEdit->toPlainText().size() != 0) {
		_ui.packingQComboBox->setEnabled(true);
	} else {
		_ui.packingQComboBox->setEnabled(false);
	}

	if(_ui.textQTextEdit->hasFocus() == true) {
		_originalText = _ui.textQTextEdit->toPlainText();
		_lastPackingIndex = 0;
		_ui.packingQComboBox->setCurrentIndex(0);
	}

	// Update the status bar and checks if the user input is valid
	_updateStatus();

}

void kms::MainWindow::slot_accountQComboBoxCurrentIndexChanged(int index) {
index = index;
	// Gets the right account from the account list using the combo box value as index
	QString value = _ui.accountQComboBox->currentText();
	for(QList<Account *>::iterator i = _accountList.begin(); i != _accountList.end(); i++) {
		if((*i)->getName() == value) {
			_currentAccount = (*i);
			Provider *provider = _currentAccount->getProvider();
			_ui.senderQLabel->setVisible(provider->acceptsSender());
			_ui.senderQLineEdit->setVisible(provider->acceptsSender());
			_ui.recipientQLabel->setVisible(provider->acceptsRecipient());
			_ui.recipientTelephoneNumber->setVisible(provider->acceptsRecipient());
			_ui.subjectQLabel->setVisible(provider->acceptsSubject());
			_ui.subjectQLineEdit->setVisible(provider->acceptsSubject());
			_ui.usedCharactersQLabel->setVisible(provider->acceptsText());
			_ui.textQLabel->setVisible(provider->acceptsText());
			_ui.textQTextEdit->setVisible(provider->acceptsText());
			break;
		}
	}

	// Updates the status bar and checks if the user input is valid
	_updateStatus();

	return;

}

void kms::MainWindow::slot_packingQComboBoxCurrentIndexChanged(int index) {

	QPalette qPalette = _ui.textQTextEdit->palette();

	if(index == 0) {
		QPalette qPalette = _ui.textQTextEdit->palette();
		qPalette.setColor(QPalette::Base,QColor(255,255,255));
		_ui.textQTextEdit->setPalette(qPalette);
		if(_lastPackingIndex != 0) {
			_ui.textQTextEdit->setText(_originalText);
		}
	} else {
		qPalette.setColor(QPalette::Base,QColor(220,220,220));
		_ui.textQTextEdit->setPalette(qPalette);
	}

	if(_lastPackingIndex == 0) {
		_originalText = _ui.textQTextEdit->toPlainText();
	}

	switch(index) {
		// Simple
		case 1:
			_ui.textQTextEdit->setPlainText(_textPacker.pack(_originalText, &_simpleAlgorithm));
			break;
		// Elegant
		case 2:
			_ui.textQTextEdit->setPlainText(_textPacker.pack(_originalText, &_elegantAlgorithm));
			break;
		// Smart
		case 3:
			_ui.textQTextEdit->setPlainText(_textPacker.pack(_originalText, &_smartAlgorithm));
			break;
		// Compact
		case 4:
			_ui.textQTextEdit->setPlainText(_textPacker.pack(_originalText, &_compactAlgorithm));
			break;
		// Cryptic
		case 5:
			_ui.textQTextEdit->setPlainText(_textPacker.pack(_originalText, &_crypticAlgorithm));
			break;
		// Crazy
		case 6:
			_ui.textQTextEdit->setPlainText(_textPacker.pack(_originalText, &_crazyAlgorithm));
			break;
		// Nasty
		case 7:
			_ui.textQTextEdit->setPlainText(_textPacker.pack(_originalText, &_nastyAlgorithm));
			break;
		default:
			break;
	}

	_lastPackingIndex = index;

}

void kms::MainWindow::slot_useAsReceiverQPushButtonClicked() {

	QString recipients;

	int count = 0;
	QTableWidgetItem *item2 = 0;

	for(int row = 0; row < _ui.indexBookTableWidget->rowCount(); row++) {
		QTableWidgetItem *item0 = _ui.indexBookTableWidget->item(row, 0);
	        if(item0->checkState() == Qt::Checked) {
			QTableWidgetItem *item1 = _ui.indexBookTableWidget->item(row, 1);
			recipients.append(", " + item1->text());
			item0->setCheckState(Qt::Unchecked);
			count++;
			item2 = _ui.indexBookTableWidget->item(row, 2);
		}
	}

	recipients.remove(0,2);

	if(count == 1) {
		if(item2->text() == QApplication::translate("MainWindow", "Generic (tin.it)", 0, QApplication::UnicodeUTF8) ) {
			_ui.accountQComboBox->setCurrentIndex(0);
		}
		if(item2->text() == QApplication::translate("MainWindow", "Generic (vodafone.it MMS)", 0, QApplication::UnicodeUTF8) ) {
			_ui.accountQComboBox->setCurrentIndex(1);
		}
		if(item2->text() == QApplication::translate("MainWindow", "Generic (vodafone.it SMS)", 0, QApplication::UnicodeUTF8) ) {
			_ui.accountQComboBox->setCurrentIndex(2);
		}
	}

	_ui.recipientTelephoneNumber->getTelephoneNumberQLineEdit()->setText(recipients);
	_ui.qTabWidget->setCurrentIndex(0);

}

void kms::MainWindow::slot_newQPushButtonClicked() {

	int row = _ui.indexBookTableWidget->rowCount();
	_ui.indexBookTableWidget->setRowCount(row + 1);
	QTableWidgetItem *item0 = new QTableWidgetItem("");
	QTableWidgetItem *item1 = new QTableWidgetItem("");
	QTableWidgetItem *item2 = new QTableWidgetItem(QApplication::translate("MainWindow", "<None>", 0, QApplication::UnicodeUTF8));
	item0->setCheckState(Qt::Unchecked);
	_ui.indexBookTableWidget->setItem(row, 2, item2);
	_ui.indexBookTableWidget->setItem(row, 1, item1);
	_ui.indexBookTableWidget->setItem(row, 0, item0);
	_ui.indexBookTableWidget->openPersistentEditor(item2);

}

void kms::MainWindow::slot_removeQPushButtonClicked() {

	for(int row = 0, k = 0; row-k < _ui.indexBookTableWidget->rowCount(); row++) {
		QTableWidgetItem *item0 = _ui.indexBookTableWidget->item(row-k, 0);
	        if(item0->checkState() == Qt::Checked) {
			_ui.indexBookTableWidget->removeRow(row-k);
			k++;
		}
	}

	slot_indexBookTableWidgetCellChanged(0, 0);

}

void kms::MainWindow::slot_indexBookTableWidgetCellChanged(int row, int column) {

	row = row;

	if(column != 0) {
		return;
	} else {
		int count = 0;
		for(int row = 0; row < _ui.indexBookTableWidget->rowCount(); row++) {
			QTableWidgetItem *item0 = _ui.indexBookTableWidget->item(row, 0);
		        if(item0->checkState() == Qt::Checked) {
				_ui.useAsReceiverQPushButton->setEnabled(true);
				_ui.removeQPushButton->setEnabled(true);
				count++;
			}
		}
		if(count==0) {
			_ui.useAsReceiverQPushButton->setEnabled(false);
			_ui.removeQPushButton->setEnabled(false);
		}
	}

}

void kms::MainWindow::slot_useAsTextQPushButtonClicked() {

	QListWidgetItem *currentItem = _ui.modelsQListWidget->currentItem();
	QString text = currentItem->text();

	QString messageText = _ui.textQTextEdit->toPlainText();
	if(messageText.size() == 0 || QString::compare(text, messageText, Qt::CaseSensitive) == 0) {
		if(messageText.size() == 0) {
			if(currentItem) {
				_ui.textQTextEdit->setText(text);
			}
			_ui.qTabWidget->setCurrentIndex(0);
		}
	} else {
		QMessageBox::StandardButton reply;
		reply = QMessageBox::question(this, tr("Use as text"), "The text of the message is not empty, can i overwrite it?", QMessageBox::Yes | QMessageBox::Cancel);
		if(reply == QMessageBox::Yes) {
			if(currentItem) {
				_ui.textQTextEdit->setText(text);
				_ui.qTabWidget->setCurrentIndex(0);
			}
		} else if (reply == QMessageBox::No) {
			// Does nothing...
		}
	}

}

void kms::MainWindow::slot_newModelQPushButtonClicked() {

	bool okFlag;
	QString text = QInputDialog::getText(this, tr("New"), tr("Create a new model:"), QLineEdit::Normal, "", &okFlag);
	if(okFlag && !text.isEmpty()) {
		_ui.modelsQListWidget->addItem(text);
	}

}

void kms::MainWindow::slot_removeModelQPushButtonClicked() {

	QListWidgetItem *currentItem = _ui.modelsQListWidget->currentItem();
	if(currentItem) {
		int row = _ui.modelsQListWidget->row(currentItem);
		_ui.modelsQListWidget->takeItem(row);
		delete currentItem;
		if(_ui.modelsQListWidget->count() > 0) {
			_ui.modelsQListWidget->setCurrentRow(0);
		}
	}

	slot_modelsQListWidgetItemSelectionChanged();

}

void kms::MainWindow::slot_modelsQListWidgetItemDoubleClicked(QListWidgetItem *item) {

	QString oldText = item->text();
	bool okFlag;
	QString headString = "Modify the following model:\n\n" + oldText + "";
	QString text = QInputDialog::getText(this, tr("Modify"), tr(headString.toAscii()), QLineEdit::Normal, oldText, &okFlag);
	if(okFlag && !text.isEmpty()) {
		item->setText(text);
	}

}

void kms::MainWindow::slot_modelsQListWidgetItemSelectionChanged() {

	int count = _ui.modelsQListWidget->count();
	QListWidgetItem *currentItem = _ui.modelsQListWidget->currentItem();
	if(currentItem && count > 0) {
		_ui.useAsTextQPushButton->setEnabled(true);
		_ui.removeModelQPushButton->setEnabled(true);
	} else {
		_ui.useAsTextQPushButton->setEnabled(false);
		_ui.removeModelQPushButton->setEnabled(false);
	}

}

void kms::MainWindow::slot_sendQPushButtonClicked() {

	// Get the credentials from the current account
	QString username = _currentAccount->getUsername();
	QString password = _currentAccount->getPassword();

	// Get some informations from the user interface
	QString sender = _ui.senderQLineEdit->text();
	QString recipient = _ui.recipientTelephoneNumber->getTelephoneNumberQLineEdit()->text();
	QString subject = _ui.subjectQLineEdit->text();
	QString text = _ui.textQTextEdit->toPlainText();

	// Send the message with the current account
//unused	Provider *provider = _currentAccount->getProvider();
	_ui.sendQPushButton->setVisible(false);
	QStringList recipientList = recipient.split(QRegExp(" *, *"));
	QString s;
	QString isare;
	if(recipientList.size() == 1) {
		s = "";
		isare = "is";
	} else {
		s = "s";
		isare = "are";
	}
	_ui.qStatusBar->showMessage("The message" + s + " " + isare + " being sending...");
	for(int i = 0; i < recipientList.size(); i++) {
		QString currentRecipient = recipientList[i].trimmed();
		_scheduler->enqueueJob(Job(_currentAccount, Message(sender, currentRecipient, subject, text)));
		// Add an item to scheduler
		QStandardItemModel *model = (QStandardItemModel *) _ui.schedulerQTreeView->model();
		model->insertRow(0);
		model->setData(model->index(0, 0), QDateTime::currentDateTime());
		model->setData(model->index(0, 1), _currentAccount->getName());
		model->setData(model->index(0, 2), currentRecipient);
		model->setData(model->index(0, 3), text);
		model->setData(model->index(0, 4), "Sent");
	}
	_ui.sendQPushButton->setVisible(true);
	_ui.qStatusBar->showMessage("The message" + s + " should have been sent.", 5000);

}

void kms::MainWindow::slot_showAccountInfoStateChanged(int status) {		// TO REMOVE

	if(status == Qt::Checked) {
// OLD!		_ui.accountInfoQLabel->setVisible(true);
	}
	if(status == Qt::Unchecked) {
// OLD!		_ui.accountInfoQLabel->setVisible(false);
	}

}
